/********************************************************************/
/********************************************************************/
/*** Makro          F1_LD_F2                                      ***/
/***                                                              ***/
/*** Die Variable VAR enth"alt die Werte                          ***/
/*** Die Faktoren TIME1 und TIME2 muessen nicht notwendig         ***/
/*** Zeitpunkte enthalten                                         ***/
/*** der Faktor SUBJECT soll die Patienten/ Probanden enthalten   ***/
/*** QUELLE: - DFG, Br 655/11-1                                   ***/
/***         - Akritas, Brunner Langer ....                       ***/
/*** geschrieben von Andreas Oelerich, AMS Goettingen, FEB  1998  ***/
/*** berarbeitet von Sebastian Domhof, September 2001            ***/
/*** berarbeitet von Sebastian Domhof, Juni 2000                 ***/
/*** (Anpassung an SAS 8.0)                                       ***/
/********************************************************************/
%macro f1_ld_f2 (data    =,
                 var     =,
                 factor  =,
                 TIME1   =,
                 TIME2   =,
                 SUBJECT = );
/********************************************************************/
%MACRO middle(dat=,var=,fak=,pat=,t1=,t2=);
 PROC SORT DATA=&dat;
  BY &t1 &t2 &pat &fak;

 PROC RANK DATA = &dat TIES = mean out = _daten_;
  VAR &var;
  RANKS rijk;

 PROC FREQ DATA = _daten_;
  TABLES &fak*&pat*&t1*&t2 / OUT = _ni_ NOPRINT;

 PROC FREQ DATA = _ni_;
  TABLES &fak*&pat / OUT=ni NOPRINT;

 PROC SORT DATA = _daten_;
  BY &fak &pat &t1 &t2;

 PROC MEANS DATA = _daten_ NOPRINT;
  BY &fak &pat &t1 &t2 ;
  VAR rijk;
  OUTPUT OUT = _value_ N = mijk MEAN = RmIJK_;

 DATA _value_;
  SET _value_;
  KEEP mIJK RmIJK_ &fak &pat &t1 &t2;
 RUN;
%MEND;

%middle(dat=&data,var=&var,fak=&factor,pat=&subject,t1=&time1,t2=&time2);

proc iml worksize=120;

%LET faktora = &factor;
%LET faktorb = &TIME1;
%LET faktorc = &TIME2;
%LET faktord = &SUBJECT;

reset nolog;
RESET LINESIZE=80;
start wald(NN,C,V,p,df,Q,p_val);
Q     = NN * (C * p)` * (ginv(C * V * C`)) * (C * p);
if Q >= 0 then p_val = 1 - probchi(Q,df);
else p_val = .;
finish;

START box(N,p_dach,C,V_d,B1,fg,p);
T        = C` * GINV(C * C`) * C;
tv       = T * V_d;
B1       = (N * p_dach` * T * p_dach) / (trace(tv));
fg       = ((trace(tv))**2)/(trace(tv * tv));
fg2      = 10000;
if B1 >= 0 & fg > 0 then p = 1 - probf(B1,fg,fg2);
else p = .;
FINISH;
START cov (NN,RIJK_,ni,d,V);
/***********************************************************************
    Die Funktion cov berechnet die Kovarianzmatrix des

                          general mixed models

    Die vorliegende Funktion berechnet die Kovarianz des multivariaten
    Modells, Spezialflle wie das CS-Modell mssen extra programmiert
    werden.

    QUELLE: Vorlesungsskript
    'Theory of rank Tests in factorial designs' von Prof. E. Brunner
    15. Oktober 1997.
    Programmiert von:       Andreas Oelerich, AMS Gttingen, Januar 1998

    Die Parameter sind
    NN    : Anzahl der observation
    RIJK_ : die Rnge wobei ber die Messwiederholung gemittelt ist
            (MACRO middle!)
    ni    : Der Vektor der die Anzahl der subjects enthlt
    d     : Anzahl der abhngigen Levels (repeated measures)
    V     : Die Kovarianz die zurckgegeben wird
***********************************************************************/
n = SUM(ni);                        /* Anzahl der subjects            */
r = NROW(ni);                       /* Anzahl der unabhngigen Levels */
pijk_ = 1/NN * (RIJK_ - 1/2);       /* Die im Vektor RIJK_ enthaltenen
                                       Rnge sollen die Mittelrnge
                                       ber die Messwiederholungen der
                                       subjects sein -> MACRO middle!
                                       pijk_ sind die Phi_ijk. des
                                       Papers                         */
IND = LOC(pijk_=.);                 /* wo sind missings?              */
IF NCOL(IND) >0 THEN pijk_[IND] = 0;
pijk_ = SHAPE(pijk_,n,d);           /* ... als Matrix                 */
lijk  = J(n,d,1);                   /* die lambda_ijk des Papers      */
IF NCOL(IND) > 0 THEN lijk[IND] = 0;/* Kodierung !                    */
lp    = lijk # pijk_ ;
DO i=1 TO r;
   IF i=1 THEN DO; r1 = 1;                  r2 = ni[i];        END;
          ELSE DO; r1 = SUM(ni[1:i-1]) + 1; r2 = SUM(ni[1:i]); END;
   lij_   = (lijk[r1:r2,])[+,];     /* ist Vektor!!                   */
   l_1lp  = lp[r1:r2,];
   pij__  = (1 / lij_) # (l_1lp[+,]);
                                    /* Phi_ij..                       */
   vij_d  = J(d,d,0);
   vijj_d = J(d,d,0);
   DO j=1 TO d;
      pij_h = J(ni[i],1,1) @ pij__[j]; /* pij_h ist eine Hilfsmatrix
                                       die die Mittel (pij__)
                                       aufplustert, so ist die
                                       Operation "-" unten zulssig
                                       wegen gleicher Dimension       */
      vij_d[j,j] = SUM(( (lijk[r1:r2,j]) # (pijk_[r1:r2,j] - pij_h))##2) ;
      faktor2    = ni[i] / (lij_[j] * ( lij_[j] - 1 ) );
      vij_d[j,j] = vij_d[j,j] * faktor2;
                                    /* vij_d sind die Schtzer der
                                       Diagonalelemente der Kovarianz-
                                       matrix Vn                      */
      DO j2=1 TO d;
         IF j2 ^= j
            THEN DO;
                   pij_h2 = J(ni[i],1,1) @ pij__[j2];
                   lijj_        = SUM(lijk[r1:r2,j] # lijk[r1:r2,j2]);
                   vijj_d[j,j2] = SUM((lijk[r1:r2,j]) # (lijk[r1:r2,j2])
                    # (pijk_[r1:r2,j] - pij_h) # (pijk_[r1:r2,j2] - pij_h2) );
                   faktor       = ni[i] / ((lij_[j]-1) * (lij_[j2]-1)
                                  + lijj_ - 1);
                   vijj_d[j,j2] = faktor * vijj_d[j,j2];
                 END;               /* of THEN DO                     */
      END;                          /* of DO : j2                     */
   END;                             /* of DO : j                      */
   Vi = n / ni[i] * (vij_d + vijj_d);
   IF i=1 THEN V = Vi;
          ELSE V = BLOCK(V,Vi);
END;                                /* of DO : i                      */
FINISH;                             /* of FUNCTION : cov              */



/*********************************************************************
   Einlesen der Rangmittel und des Designs
*********************************************************************/
USE _value_;
READ ALL VAR{mijk}     INTO m_ijkl;
READ ALL VAR{RmIJK_}   INTO RmIJKl;
READ ALL VAR{&factor}  INTO beh_;
READ ALL VAR{&subject} INTO pat_;
READ ALL VAR{&time1}   INTO t1_;
READ ALL VAR{&time2}   INTO t2_;
CLOSE _value_;
ranks = RmIJKl;
lev_a = unique(beh_);                 /* Die Stufen des Faktors A ***/
lev_b = unique(t1_);                  /* Die Stufen des Faktors B ***/
lev_c = UNIQUE(t2_);                  /* Die Stufen des Faktors C ***/
lev_d = unique(pat_);                 /* Die Stufen des Faktors D ***/
a     = ncol(lev_a);                  /* Anzahl der Stufen von A  ***/
b     = ncol(lev_b);                  /* Anzahl der Stufen von B  ***/
c     = ncol(lev_c);                  /* Anzahl der Stufen von C  ***/
n     = ncol(lev_d);                  /* Anzahl der subjects      ***/
IF n*b*c ^= nrow(m_ijkl)
  THEN DO;
       print 'Error',
             'If you have missing values (empty cell(s))',
             'please fill in one observation with a .';
       END;
m_ijkl = SHAPE(m_ijkl,n,b*c);
NN    = sum(m_ijkl);                  /* Anzahl der Daten           */
n_i = J(NCOL(lev_a),1,0);

DO i=1 TO a;
   IND    = LOC(beh_=lev_a[i]);
   n_i[i] = NCOL(UNIQUE(pat_[IND]));
END;
/********************************************************************/
/*** i ist der Laufindex des Faktors A  i=1,...,na                ***/
/*** j ist der Laufindex des Faktors B  j=1,...,nb                ***/
/*** k ist der Laufindex des Faktors C  k=1,...,nc                ***/
/*** l ist der Laufindex der Subjects   l=1,...,nd                ***/
/*** s ist der Laufindex der Messwiederholungen s=1,...,m_ijkl    ***/
/********************************************************************/
/********************************************************************/
/*** Definition der Contrastmatrizen                              ***/
/********************************************************************/
PA    = I(a) - j(a,a,1/a);
PB    = I(b) - j(b,b,1/b);
PC    = i(c) - j(c,c,1/c);
A1    = j(1,a,1/a);
B1    = j(1,b,1/b);
C1    = j(1,c,1/c);
CA    = PA   @ B1   @ C1;
CB    = A1   @ PB   @ C1;
CC    = A1   @ B1   @ PC;
CAB   = PA   @ PB   @ C1;
CAC   = PA   @ B1   @ PC;
CBC   = A1   @ PB   @ PC;
CABC  = PA   @ PB   @ PC;
CA_BC = PA   @ i(b) @ i(c);
CB_AC = i(a) @ PB   @ i(c);
CC_AB = i(a) @ i(b) @ PC;
/********************************************************************/
/*** Rijkl    = Vector of rank means for cell (i,j,k)             ***/
/*** Rijkl_ma = (anxbc)-Matrix der Rangmittelwerte                ***/
/*** ranks    = R_ijks                                            ***/
/*** p_d      = estimate for relativ treatment effect             ***/
/********************************************************************/
p       = (ranks - 1/2) / NN ;
r_ijkl  = ranks;       /* Matrix der Rangmittel ueber die Mess-     */
                       /* wiederholungen                            */
r_ijkl  = shape(r_ijkl,n,b*c);
/********************************************************************/
/*** In der Matrix r_ijk stehen die Rangmittel der Patienten      ***/
/********************************************************************/
r_ijk  = j(a,b*c,0);
m_ijk  = j(a,b*c,0);
do i=1 to a;
   if i>1 then do; r1 = sum(n_i[1:i-1]) + 1; end;
          else do; r1 = 1;                   end;
   r2 = sum(n_i[1:i]);
   r_ijk[i,] = (r_ijkl[r1:r2,])[:,];
   m_ijk[i,] = (m_ijkl[r1:r2,])[+,];
end;
p_s = 1/NN *(r_ijk - 1/2);
p_d = shape(p_s,a*b*c,1);

RUN COV(NN,ranks,n_i,b*c,V);
V = V * NN / n ;
/********************************************************************/
/*** Die Teststatistiken                                          ***/
/*** vgl. Agresti & Brunner S. 21                                 ***/
/********************************************************************/
Q     = j(10,1,0);
p_val = j(10,1,0);
df    = j(10,1,0);
df[1] = a-1;
df[2] = b-1;
df[3] = c-1;
df[4] = (a-1) * (b-1);
df[5] = (a-1) * (c-1);
df[6] = (b-1) * (c-1);
df[7] = (a-1) * (b-1) * (c-1);
df[8] = (a-1) * b * c ;
df[9] = a * (b-1) * c ;
df[10]= a * b * (c-1);
/*********************************************************************
     Wald-Typ Statistiken
*********************************************************************/
run wald(NN,CA   ,V,p_d,df[1] ,QF,p_v) ;            /* Effekt A     */
Q[1]  = QF; p_val[1]  = p_v;
run wald(NN,CB   ,V,p_d,df[2] ,QF,p_v) ;            /* Effekt B     */
Q[2]  = QF; p_val[2]  = p_v;
run wald(NN,CC   ,V,p_d,df[3] ,QF,p_v) ;            /* Effekt C     */
Q[3]  = QF; p_val[3]  = p_v;
run wald(NN,CAB  ,V,p_d,df[4] ,QF,p_v) ;            /* Effekt AB    */
Q[4]  = QF; p_val[4]  = p_v;
run wald(NN,CAC  ,V,p_d,df[5] ,QF,p_v) ;            /* Effekt AC    */
Q[5]  = QF; p_val[5]  = p_v;
run wald(NN,CBC  ,V,p_d,df[6] ,QF,p_v) ;            /* Effekt BC    */
Q[6]  = QF; p_val[6]  = p_v;
run wald(NN,CABC ,V,p_d,df[7] ,QF,p_v) ;            /* Effekt ABC   */
Q[7]  = QF; p_val[7]  = p_v;
run wald(NN,CA_BC,V,p_d,df[8] ,QF,p_v) ;            /* Effekt A_BC  */
Q[8]  = QF; p_val[8]  = p_v;
run wald(NN,CB_AC,V,p_d,df[9] ,QF,p_v) ;            /* Effekt B_AC  */
Q[9]  = QF; p_val[9]  = p_v;
run wald(NN,CC_AB,V,p_d,df[10],QF,p_v) ;            /* Effekt C_AB  */
Q[10] = QF; p_val[10] = p_v;

test = Q || df || p_val;
/*********************************************************************
   Die BOX-Approximation
*********************************************************************/
Q_box = j(10,1,0);
p_box = j(10,1,0);
dfbox = j(10,1,0);
RUN box(NN,p_d,CA   ,V,QF_,DF_,p_);             /* Effekt A         */
Q_box[1]  = QF_; dfbox[1]  = DF_ ; p_box[1]  = p_;
RUN box(NN,p_d,CB   ,V,QF_,DF_,p_);             /* Effekt B         */
Q_box[2]  = QF_; dfbox[2]  = DF_ ; p_box[2]  = p_;
RUN box(NN,p_d,CC   ,V,QF_,DF_,p_);             /* Effekt C         */
Q_box[3]  = QF_; dfbox[3]  = DF_ ; p_box[3]  = p_;
RUN box(NN,p_d,CAB  ,V,QF_,DF_,p_);             /* Effekt AB        */
Q_box[4]  = QF_; dfbox[4]  = DF_ ; p_box[4]  = p_;
RUN box(NN,p_d,CAC  ,V,QF_,DF_,p_);             /* Effekt AC        */
Q_box[5]  = QF_; dfbox[5]  = DF_ ; p_box[5]  = p_;
RUN box(NN,p_d,CBC  ,V,QF_,DF_,p_);             /* Effekt BC        */
Q_box[6]  = QF_; dfbox[6]  = DF_ ; p_box[6]  = p_;
RUN box(NN,p_d,CABC ,V,QF_,DF_,p_);             /* Effekt ABC       */
Q_box[7]  = QF_; dfbox[7]  = DF_ ; p_box[7]  = p_;
RUN BOX(NN,p_d,CA_BC,V,QF_,DF_,p_);             /* Effekt A_BC      */
Q_box[8]  = QF_; dfbox[8]  = DF_ ; p_box[8]  = p_;
RUN BOX(NN,p_d,CB_AC,V,QF_,DF_,p_);             /* Effekt B_AC      */
Q_box[9]  = QF_; dfbox[9]  = DF_ ; p_box[9]  = p_;
RUN BOX(NN,p_d,CC_AB,V,QF_,DF_,p_);             /* Effekt C_AB      */
Q_box[10] = QF_; dfbox[10]  = DF_ ; p_box[10]  = p_;

ANOVA = Q_box || Dfbox || p_box ;
/********************************************************************/
start nu_char(v,v_neu) ;
if type (v) ='N' then v_neu = char(v,max(int (log10(max(abs(v))))+1));
 else v_neu = trim(v);
finish;
/* Ende der Funktion nu_char                                       **/
/********************************************************************/
/********************************************************************/
start CLI(a,b,c,d,n,m_ijk);
 reset center;
 class  = {A &faktora, T &faktorb, C &faktorc};
 levels = a // b // c ;
 *print / ;
 print 'F1_LD_F2 --- subjects(A) x T x C',
       'A, T, C: fixed, subjects: random';
 print "SAS-datafile-name: &data",
       "Response variable: &var";
 print 'Class Level Information' ;
 print class levels;
 reset noname;
 print 'Total number of observations ' (sum(m_ijk)) ' ',
       'Total number of subjects     ' n ' ',
       'Number of missing values     ' (NCOL(LOC(m_ijk=0)))  ;
finish ; /* Ende von O_CLI */;
/********************************************************************/
/********************************************************************/
/*** Diese Funktion schreibt die Faktorstufenkombinationen zweier   */
/*** Faktoren in den Vektor help.                                   */
/********************************************************************/
start out1(ch_1,ch_2,ch1_n,ch2_n,help);
do i=1 to ch1_n;
   do j=1 to ch2_n;
      help = help // (ch_1[i] + '*' + ch_2[j]);
   end;
end;
finish; /* out1                                                     */
/********************************************************************/
/*** Diese Funktion schreibt die Faktorstufenkombinationen dreier   */
/*** Faktoren in den Vektor help                                    */
start out2(ch_1,ch_2,ch_3,ch1_n,ch2_n,ch3_n,help);
 do i=1 to ch1_n;
    do j=1 to ch2_n;
       do k=1 to ch3_n;
          help = help // (ch_1[i] + '*' + ch_2[j] + '*' + ch_3[k]);
       end;
    end;
 end;
finish; /* out2                                                     */
/********************************************************************/
/*** Diese Funktion erstellt den source-Vektor der fuer den Output  */
/*** benoetigt wird.                                                */
/********************************************************************/
start source_k(cha_1,cha_2,cha_3,source) global (a,b,c);
%let name1 = %trim(&FAKTORA);
%let name2 = %trim(&FAKTORB);
%let name3 = %trim(&FAKTORC);
spalte1 =j(a,1,"&name1") // j(b,1,"&name2") // j(c,1,"&name3")
                         // j(a*b,1,"&name1"+'*'+"&name2"+' ')
                         // j(a*c,1,"&name1"+'*'+"&name3"+' ')
                         // j(b*c,1,"&name3"+'*'+"&name2"+' ')
                         // j(a*b*c,1,"&name1"+'*'+"&name2"+'*'+"&name3"+' ');
run out1(cha_1,cha_2,a,b,text);
run out1(cha_1,cha_3,a,c,text);
run out1(cha_3,cha_2,c,b,text);
run out2(cha_1,cha_2,cha_3,a,b,c,text);
spalte2 = (cha_1`) // (cha_2`) // (cha_3`) // text;
source  = concat (spalte1,spalte2);
finish; /* source_k                                                 */
/********************************************************************/
/********************************************************************/
/*** Ausgabe der rankmeans usw.                                  ****/
/********************************************************************/
start rmeans(a,b,c,source,mijk,p_d,r_ijk);
m_ijk = shape(mijk,a*b*c,1);
p_ijk = p_d;
r_ijk = shape(r_ijk,a*b*c,1);
*print p_ijk;
f_a   = mijk[,+];
CB    = j(1,a*c,1) @ i(b);
CC    = j(1,a,1) @ i(c) @ j(1,b,1) ;
CAB   = i(a) @ j(1,c,1) @ i(b);
CAC   = i(a) @ i(c) @ j(1,b,1);
f_b   = CB * m_ijk;
f_c   = CC * m_ijk;
f_ab  = CAB * m_ijk;
f_ac  = CAC * m_ijk;
f_bc  = mijk[+,]`;
f_abc = shape(mijk,a*b*c,1);
nor   = f_a // f_b // f_c // f_ab // f_ac // f_bc // f_abc;
p_a   = (shape(p_ijk,a,b*c))[,:];
p_b   = CB  * p_ijk / (a*c);
p_c   = CC  * p_ijk / (a*b);
p_ab  = CAB * p_ijk / c;
p_ac  = CAC * p_ijk / b;
p_bc  = ((shape(p_ijk,a,b*c))[:,])`;
p_abc = p_ijk;
p     = p_a // p_b // p_c // p_ab // p_ac // p_bc // p_abc ;
r_a   = (shape(r_ijk,a,b*c))[,:];
r_b   = CB  * r_ijk / (a*c);
r_c   = CC  * r_ijk / (a*b);
r_ab  = CAB * r_ijk / c;
r_ac  = CAC * r_ijk / b;
r_bc  = ((shape(r_ijk,a,b*c))[:,])`;
r_abc = r_ijk;
r_vec = r_a // r_b // r_c // r_ab // r_ac // r_bc // r_abc ;
*print /;
print 'RTE  = Relative Treatment Effects          ',
      'Nobs = Number of observations (do not count',
      'the repeated measurements within the cells)';
print source[c={source}] r_vec[format=6.5 c={"Rank mean"}]
      nor[c={"Nobs"}] p[format=6.5 c={"RTE"}];
finish;
/*******************************************************************/
/*** Ausgabe der Warnungen                                       ***/
/*******************************************************************/
start warning(b,c,n,V,m_ijk);
if b*c > n | det(V) = 0 | min(eigval(V)) < 0 then
  print / '';
reset nocenter;
if b*c > n then
  print 'Warning:',
        'There are less subjects than sub-plot factor level combinations.';
else if det(V)=0 then
  print 'Warning:',
        'Do not use the Wald-type-statistic, because the covariance matrix is singular.';
if min(eigval(V)) < 0 & NCOL(LOC(m_ijk=0)) > 0 then
  print 'Warning:',
        'The estimated covariance matrix is not positive semidefinite due to missing values.';
reset center;
finish;
/********************************************************************/
/*** Ausgabe der Quadratformen und der p_values                  ****/
/********************************************************************/
/*         Im Output heisst der erste Zeitfaktor T !                */
/********************************************************************/
start test_out(wald,anova);
source = {"A" , "T" , "C" , "AT" , "AC" , "TC" , "ATC" ,
          "A|TC" , "T|AC" , "C|AT"};
row    =  {"A" , "T" , "C" , "AT" , "AC" , "TC" , "ATC"};
col1   = {"QF"  "DF"  "p-value"};
wald   = wald[1:7,];
anova  = anova[1:7,];
print /;
print 'Wald-type-statistic',
      'Approximation for large sample sizes with Chi^2_DF';
print wald[format=6.5][r=row][c=col1];
print 'ANOVA-Type-statistic',
      'Box-Approximation for small sample size with Chi^2_(df)',
      '-estimated degree of freedom';
print anova[format=6.5][r=row][c=col1];
finish;
/********************************************************************/
/********************************************************************/
dataname = name(&data);
varname  = name(&var);
fa1_name = name(&faktora);
fa2_name = name(&faktorb);
fa3_name = name(&faktorc);
fa4_name = name(&faktord);
run nu_char(lev_a, leva);
run nu_char(lev_b, levb);
run nu_char(lev_c, levc);
run nu_char(lev_d, levd);
run cli(a,b,c,d,n,m_ijkl);
run source_k(leva,levb,levc,source);
run rmeans(a,b,c,source,m_ijk,p_d,r_ijk);
run warning(b,c,n,V,m_ijk);
run test_out(test,anova);
quit;
%mend f1_ld_f2;
