/*********************************************************************
   factor C, zuf"allig, nsum Stufen, nsum subjects (Index k)
   factor B, fest,       c Stufen,  c Zeitpunkte   (Index j)
   B und C sind vollst"andig gekreuzt
   X_jk verteilt nach F_j
   Datum: 28. Januar 1998
   Quelle: Brunner, Munzel, Puri (1996)
   geschrieben von Steffen Ballerstedt und
                   Andreas Oelerich;  AMS Gttingen, Feb. 1998
   berarbeitet von Sebastian Domhof, September 2001
   berarbeitet von Sebastian Domhof, Juni 2000 (Anpassung an SAS 8.0)
   letzte nderung: Sebastian Domhof, 26.06.2000
*********************************************************************/
/********************************************************************/
%macro LD_F1(DATA= ,VAR = ,TIME = ,SUBJECT = ,
       DATA_PT = _no_ ,VAR_PT = _no_ ,TIME_PT = _no_ );
/********************************************************************/

%MACRO middle(dat=,var=,pat=,t=);

 PROC SORT DATA=&dat;
  BY &t &pat;

 PROC RANK DATA = &dat TIES = mean out = _daten_;
  VAR &var;
  RANKS rijk;

 PROC FREQ DATA = _daten_;
  TABLES &pat*&t / OUT = _ni_ NOPRINT;

 PROC FREQ DATA = &dat;
  TABLES &pat / OUT=_nipat_ NOPRINT;

 PROC FREQ DATA = &dat;
  TABLES &t / OUT=_nitime_ NOPRINT;

 PROC SORT DATA = _daten_;
  BY &pat &t;

 PROC MEANS DATA = _daten_ NOPRINT;
  BY &pat &t ;
  VAR rijk;
  OUTPUT OUT = _value_ N = mijk MEAN = RmIJK_;

 DATA _value_;
  SET _value_;
  KEEP mIJK RmIJK_ &pat &t;
 RUN;

%MEND;

%middle(dat=&data,var=&var,pat=&subject,t=&time);


proc iml worksize=120;
RESET LINESIZE=80;
reset nolog;
%LET factorB = &TIME;                /* Variablennamenumbenennung   */
%LET factorC = &SUBJECT;             /* dito                        */
%LET DAT_PAT = &DATA_PT;
%LET PATTERN = &VAR_PT;
%LET FAC_PAT = &TIME_PT;

/********************************************************************/
sort &DATA by &factorB &factorC;              /* richtig sortiert ! */
/********************************************************************/

/********************************************************************/
start lese_dat(b_text,c_text, x);             /* Einlesen der Daten */
 use &DATA;
 read all var{&factorB} into b_text;
 read all var{&factorC} into c_text;
 read all var{&VAR}     into X;
 close &DATA;
finish;
/********************************************************************/

/********************************************************************/
start lese_pat(dummy)                    /*   Einlesen des Patterns */
       global(weights);                  /* Diese Funktion wird nur */
 sort &DAT_PAT by &FAC_PAT;              /*   aufgerufen, wenn ein  */
 use &DAT_PAT;                           /*    Pattern existiert    */
 read all var{&PATTERN} into weights;
 close &DAT_PAT;
finish;
/********************************************************************/
START cov (NN,RIJK_,ni,d,V);
/***********************************************************************
    Die Funktion cov berechnet die Kovarianzmatrix des

                          general mixed models

    Die vorliegende Funktion berechnet die Kovarianz des multivariaten
    Modells, Spezialflle wie das CS-Modell mssen extra programmiert
    werden.

    QUELLE: Vorlesungsskript
    'Theory of rank Tests in factorial designs' von Prof. E. Brunner
    15. Oktober 1997.
    Programmiert von:       Andreas Oelerich, AMS Gttingen, Januar 1998

    Die Parameter sind
    NN    : Anzahl der observation
    RIJK_ : die Rnge wobei ber die Messwiederholung gemittelt ist
            (MACRO middle!)
    ni    : Der Vektor der die Anzahl der subjects enthlt
    d     : Anzahl der abhngigen Levels (repeated measures)
    V     : Die Kovarianz die zurckgegeben wird
***********************************************************************/
n = SUM(ni);                        /* Anzahl der subjects            */
r = NROW(ni);                       /* Anzahl der unabhngigen Levels */
pijk_ = 1/NN * (RIJK_ - 1/2);       /* Die im Vektor RIJK_ enthaltenen
                                       Rnge sollen die Mittelrnge
                                       ber die Messwiederholungen der
                                       subjects sein -> MACRO middle!
                                       pijk_ sind die Phi_ijk. des
                                       Papers                         */
IND = LOC(pijk_=.);                 /* wo sind missings?              */
IF NCOL(IND) >0 THEN pijk_[IND] = 0;
pijk_ = SHAPE(pijk_,n,d);           /* ... als Matrix                 */
lijk  = J(n,d,1);                   /* die lambda_ijk des Papers      */
IF NCOL(IND) > 0 THEN lijk[IND] = 0;/* Kodierung !                    */
lp    = lijk # pijk_ ;
DO i=1 TO r;
   IF i=1 THEN DO; r1 = 1;                  r2 = ni[i];        END;
          ELSE DO; r1 = SUM(ni[1:i-1]) + 1; r2 = SUM(ni[1:i]); END;
   lij_   = (lijk[r1:r2,])[+,];     /* ist Vektor!!                   */
   l_1lp  = lp[r1:r2,];
   pij__  = (1 / lij_) # (l_1lp[+,]);
                                    /* Phi_ij..                       */
   vij_d  = J(d,d,0);
   vijj_d = J(d,d,0);
   DO j=1 TO d;
      pij_h = J(ni[i],1,1) @ pij__[j]; /* pij_h ist eine Hilfsmatrix
                                       die die Mittel (pij__)
                                       aufplustert, so ist die
                                       Operation "-" unten zulssig
                                       wegen gleicher Dimension       */
      vij_d[j,j] = SUM(( (lijk[r1:r2,j]) # (pijk_[r1:r2,j] - pij_h))##2) ;
      faktor2    = ni[i] / (lij_[j] * ( lij_[j] - 1 ) );
      vij_d[j,j] = vij_d[j,j] * faktor2;
                                    /* vij_d sind die Schtzer der
                                       Diagonalelemente der Kovarianz-
                                       matrix Vn                      */
      DO j2=1 TO d;
         IF j2 ^= j
            THEN DO;
                   pij_h2 = J(ni[i],1,1) @ pij__[j2];
                   lijj_        = SUM(lijk[r1:r2,j] # lijk[r1:r2,j2]);
                   vijj_d[j,j2] = SUM((lijk[r1:r2,j]) # (lijk[r1:r2,j2])
                    # (pijk_[r1:r2,j] - pij_h) # (pijk_[r1:r2,j2] - pij_h2) );
                   faktor       = ni[i] / ((lij_[j]-1) * (lij_[j2]-1)
                                  + lijj_ - 1);
                   vijj_d[j,j2] = faktor * vijj_d[j,j2];
                 END;               /* of THEN DO                     */
      END;                          /* of DO : j2                     */
   END;                             /* of DO : j                      */
   Vi = n / ni[i] * (vij_d + vijj_d);
   IF i=1 THEN V = Vi;
          ELSE V = BLOCK(Vn,Vi);
END;                                /* of DO : i                      */
FINISH;                             /* of FUNCTION : cov              */


/********************************************************************/
USE _value_;
READ ALL VAR{mijk}     INTO m_ijk;
READ ALL VAR{RmIJK_}   INTO RmIJK;
READ ALL VAR{&subject} INTO pat_;
READ ALL VAR{&time}    INTO t_;
CLOSE _value_;
lev_a = unique(t_);                   /* Die Stufen des Faktors A ***/
lev_b = unique(pat_);                 /* Die Stufen des Faktors B ***/
c     = ncol(lev_a);                  /* Anzahl der Stufen von A  ***/
nsum  = ncol(lev_b);                  /* Anzahl der Subjects      ***/
IF nsum*c ^= nrow(m_ijk)
  THEN DO;
       print 'Error',
             'If you have missing values (empty cell(s))',
             'please fill in one observation with a .';
       END;
m_ijk = SHAPE(m_ijk,nsum,c);
NN    = sum(m_ijk);                   /* Anzahl der Daten           */
N     = nsum * c;
RUN COV(NN,RmIJK,nsum,c,V_d);
RmIJ_   = (SHAPE(RmIJK,nsum,c))[:,];
mvpdach = 1 / NN * (RmIJ_` - 1/2);
mvrgmat = shape(RmIJK,nsum,c);

/********************************************************************/
/*** i ist der Laufindex des Faktors A  i=1,...,a                 ***/
/*** j ist der Laufindex des Faktors B  j=1,...,b                 ***/
/*** k ist der Laufindex der Subjects   k=1,...,n                 ***/
/*** s ist der Laufindex der Messwiederholungen s=1,...,m_ijk     ***/
/********************************************************************/
/********************************************************************/
start wald_n(anz,p_dach,kontr,v_d,w,fg);
  cp  = kontr * p_dach;
  cvc = kontr * v_d * kontr`;
  w   = anz * cp` * ginv(cvc) * cp;
  fg  = trace(cvc * ginv(cvc));
finish;
/********************************************************************/

/********************************************************************/
start box1(nsum,p_dach,T,v_d,B1,fg);
 tv = T*v_d;
 B1 = (nsum*p_dach`*T*p_dach)/trace(tv);
 fg = ((trace(tv))**2)/(trace(tv*tv));
finish;  /* box1 */
/********************************************************************/

/********************************************************************/
start pattern(anz,w,p_dach,kontr,v_d,T);
  wc  = w` * kontr;
  cov = sqrt(wc * v_d * wc`);
  T   = (sqrt(anz) * wc * p_dach) / cov;
  /* T ist asymptotisch Standardnormalverteilt */
finish;  /* pattern */
/********************************************************************/

/********************************************************************/
start small(nsum,big,fg,small_w,fg2);
  small_w = (nsum - fg ) * big / (fg * (nsum - 1));
  fg2 = nsum - fg;
finish;
/********************************************************************/

/********************************************************************/
start cistzwei(X,nsum,RJKmat,RmJ_,T2,mv_);

 /* Behrens-Fisher-Problem fuer c=2: mit H_0^p(B) ------------------*/
 /* nur fuer den Spezialfall mit zwei Zeitpunkten c=2 --------------*/
 /* und auch ohne fehlende Werte -----------------------------------*/
 IND   = LOC(x=.);
 IND2  = LOC(RJKmat=.);
 bfnot = 0;
 IF NCOL(IND)>0 THEN DO;
                      bfnot = 1; RJKmat[IND2]=0;
                     END;

 /* within ranking */
IF bfnot=0 THEN DO;
                  xmat        = (shape(x,2,nsum))`;
                  RwJKmat     = j(nsum,2,0);
                  RwJKmat[,1] = ranktie(Xmat[,1]);
                  RwJKmat[,2] = ranktie(Xmat[,2]);
                END;

 /* den Varianzschaetzer s_{n,0}^2 und s_{n,p}^2 */
IF bfnot=0 THEN DO;
                  diff  = RJKmat - RwJKmat;
                  diff2 = (diff  *{-1,1}) - (RmJ_*{-1,1});
                  S_np  = ssq(diff2)/(nsum-1);
                END;

 diff3       = (RJKmat*{-1,1}) - (RmJ_*{-1,1});
 S_n0        = ssq(diff3)/(nsum-1);

 /* Berechnet die Teststatistik */
 T2          = j(2,3,.);
 T2[1,1]     = sqrt(nsum) * (RmJ_*{-1,1})/(sqrt(S_n0));

 /* Berechnung des P-Wertes */                             /* zweiseitig:        */
 T2[1,2]     = 2 * (1 - probnorm(abs(T2[1,1]))    );       /* p = 2*(1-pr(abs()) */
 T2[1,3]     = 2 * (1 - probt(abs(T2[1,1]),nsum-1));
IF bfnot=0 THEN DO;
                   T2[2,1]     = sqrt(nsum) * (RmJ_*{-1,1})/(2*sqrt(S_np));
                   T2[2,2]     = 2 * (1 - probnorm(abs(T2[2,1]))    );
                   T2[2,3]     = 2 * (1 - probt(abs(T2[2,1]),nsum-1));
                END;
mv_ = bfnot;
finish;
/********************************************************************/

/********************************************************************/
start ausgabe(tabelle,name_b,name_c,mvRmJ_,p_dach,T2,pat,NN,c,mv_,m_ijk,nsum,v_d)
      global(patt_ind,weights);

/*--- Ausgabe allgemeiner Informationen zum Design -----------------*/
 USE _nitime_; READ ALL VAR{count} INTO nJ; CLOSE _nitime_;
/*
print 'Two Factor Design --- B x C',
       'B: fixed --- C: random (subjects)';
*/
 print '  LD_1F --- subjects x T  ',
       'T: fixed, subjects: random';

/*--- Ausgabe allgemeiner Informationen zum Datensatz --------------*/

 print "SAS-data-filename: &DATA",
       "response-variable: &VAR";

 print 'Class Level Information';

 class  = {T &factorB};
 levels = c;
 print class levels;

 reset noname;

/*--- Ausgabe der Beobachtungen und der Missings -------------------*/
 ind = LOC(m_ijk=0);
 IF NCOL(IND)>0 THEN missing = NCOL(IND); ELSE missing=0;

 char_mvN = char(NN,max(int(log10(NN))+1));              /* blanc   */
 char_mis = char(missing,max(int(log10(missing+1))+1));  /* cutting */
 char_c   = char(c,max(int(log10(c))+1));                /* blanc   */
 char_n   = char(nsum,max(int(log10(nsum))+1));          /* cutting */

 print 'Total number of observations ' char_mvN ' ',
       'Total number of subjects     ' char_n   ' ',
       'Number of missing values     ' char_mis ' ' ;

 print 'RTE  = relative treatment Effects          ',
       'Nobs = Number of observations (do not count',
       'the repeated measurements within the cells)';
/*--- Ausgabe der Rangmittel und Zellbesetzungen -------------------*/

 %let fac_B = %trim(&factorB);                     /* blanc cutting */
 if type(name_b)='N'                               /* blanc cutting */
   then char_nb = name_b;     *char(name_b,max(int(log10(name_b))+1));
   else char_nb = trim(name_b);                    /* blanc cutting */

 spalte1  = j(c,1,"&fac_B");
 mvRmJ_t  = mvRmJ_`;
 char_nbt = char_nb`;
 nJ = (m_ijk[+,])`;
 print spalte1[c={'Source'}]
       char_nbt[c={"Time"}]
       mvRmJ_t[format=6.5 c={"Rank means"}]
       nJ[c={"Nobs"}]
       p_dach[format=6.5 c={"RTE"}];

/*--- Ausgabe eventueller Warnungen --------------------------------*/

if c > nsum | det(v_d) = 0 | min(eigval(v_d)) < 0 then
  print / '';
reset nocenter;
if c > nsum then
  print 'Warning:',
        'There are less subjects than sub-plot factor levels.';
else if det(v_d)=0 then
  print 'Warning:',
        'Do not use the Wald-type-statistic, because the covariance matrix is singular.';
if min(eigval(v_d)) < 0 & missing > 0 then
  print 'Warning:',
        'The estimated covariance matrix is not positive semidefinite due to missing values.';
reset center;

/*--- Ausgabe der Testergebnisse -----------------------------------*/

print /;                                                /* newpage */

/*--- Ausgabe der Statistiken fuer c=2, H_0^F und H_0^p ------------*/
 if c=2 then
  do;
   print '----------------------------------------------------------';
   print 'p-values (two-sided)';
   print T2[format = 6.5 r = {'H_0^F','H_0^p'}
            c = {'Statistics' 'N(0,1)' 't(n-1)'}];
   IF mv_ = 1 THEN
           print 'It is not possible to compute the statistic T_n^p with missing values.';
  end;

/*--- Ausgabe der Standardstatistiken fuer c>2  --------------------*/
 else do;
 zeilen  = {'Wald Chi^2(df1)',
            'Hotelling F(df1,df2)',
            'Anova Chi^2(df1)/df1'};
 spalten = {'Statistics' 'df1' 'df2' 'p-value'};

 print '----------------------------------------------------------';
 print tabelle[format=6.5 r=zeilen c=spalten];
 end; /* do */

/*--- Ausgabe des Patterns (nur, wenn Pattern existiert) -----------*/
 if patt_ind=1 then
  do;
   pat_tab = name_b // weights`;
   print '----------------------------------------------------------';
   print 'Patternd Alternatives',
         "SAS-pattern-filename: &DAT_PAT";
   print pat_tab[r={'Time points', 'Selected Pattern'}];
   print 'p-values (one-sided)';
   print pat[format=6.5 c={'Statistic' 'N(0,1)' 't(n-1)'}];
   print '----------------------------------------------------------';
  end; /* do */

finish; /* ausgabe */

/********************************************************************/

/**************** DAS HAUPTPROGRAMM *********************************/

  run lese_dat(b_text, c_text, x);

  patt_ind = 1;                      /* Als default wird der        */
                                     /* Pattern-Indikator=1 gesetzt */
  %if &DAT_PAT=_no_ %then
   %do;                              /* Liegt kein Pattern vor,     */
    patt_ind = 0;                    /* wird er hier =0 gesetzt     */
   %end;                             /* klappt nicht ohne %do, %end */

  if patt_ind=1 then run lese_pat(dummy);

  C_C    = I(c)-j(c,c,1/c);
  T_C    = C_C*ginv(C_C*C_C`)`*C_C`;

/*  Wald-Typ-Statistiken -------------------------------------------*/
  run wald_n(nsum,mvpdach,C_C,v_d,w_C,fg_C);
  if w_C >= 0 & fg_C > 0 then p_wald = 1-probchi(w_C,fg_C);
  else p_wald = .;
  wald   = w_C || fg_C || {.} ||p_wald;

/*  Hotelling-Statistik --------------------------------------------*/
  run small(nsum,w_C,fg_C,wf_C,fg2);
  if wf_C >= 0 & fg_C > 0 & fg2 > 0 then p_small = 1 - probf(wf_C, fg_C, fg2);
  else p_small = .;
  klein = wf_C || fg_C || fg2 || p_small;

/*  ANOVA-Typ-Statistiken mit Box-Approximation --------------------*/
  run box1(nsum,mvpdach,T_C,v_d,B_C2,fgb_C);
  if B_C2 >= 0 & fgb_C >0 then p_box = 1-probchi(B_C2 * fgb_c,fgb_C);
  else p_box = .;
  box   = B_C2 || fgb_C || {.} || p_box;

  tabelle = wald // klein // box;

/*  Pattern-Statistiken (nur, wenn Pattern existiert) --------------*/
 if patt_ind=1 then
  do;
   run pattern(nsum,weights,mvpdach,C_C,v_d,pat_C);
   p_patt   = 1 - probnorm(pat_C);
   fg_patt  = nsum - 1;
   p_pattkl = 1 - probt(pat_C, fg_patt);
   pat      = pat_C || p_patt || p_pattkl;
  end; /*  do  */

/* Behrens-Fisher-Statistik fuer c=2 : H_0^p(B) --(ohne missings)---*/
/*                Statistik fuer c=2 : H_0^F(B) --(ohne missings)---*/

 if c=2 then run cistzwei(x,nsum,mvrgmat,RmIJ_,T2,mv_);

/* Ausgabe der Ergebnisse ------------------------------------------*/
 run ausgabe(tabelle, lev_a,lev_b, RmIJ_,mvpdach,T2,pat,NN,c,mv_,m_ijk,nsum,v_d);
/********************************************************************/
quit; /* IML */
/********************************************************************/
%mend LD_F1;
/********************************************************************/
