/********************************************************************/
/*** Makro             LD_F2                                      ***/
/***                                                              ***/
/*** berarbeitet von Sebastian Domhof, September 2001            ***/
/*** berarbeitet von Sebastian Domhof, Juni 2000                 ***/
/*** (Anpassung an SAS 8.0)                                       ***/
/********************************************************************/
%macro ld_f2(data    =,
             var     =,
             TIME1   =,
             TIME2   =,
             SUBJECT =,);
%MACRO middle(dat=,var=,pat=,t1=,t2=);
 PROC SORT DATA=&dat;
  BY &t1 &t2 &pat;

 PROC RANK DATA = &dat TIES = mean out = _daten_;
  VAR &var;
  RANKS rijk;

 PROC FREQ DATA = _daten_;
  TABLES &pat*&t1*&t2 / OUT = _ni_ NOPRINT;

 PROC FREQ DATA = _ni_;
  TABLES &pat / OUT=ni NOPRINT;

 PROC SORT DATA = _daten_;
  BY &pat &t1 &t2;

 PROC MEANS DATA = _daten_ NOPRINT;
  BY &pat &t1 &t2 ;
  VAR rijk;
  OUTPUT OUT = _value_ N = mijk MEAN = RmIJK_;

 DATA _value_;
  SET _value_;
  KEEP mIJK RmIJK_ &pat &t1 &t2;
 RUN;
%MEND;

%middle(dat=&data,var=&var,pat=&subject,t1=&time1,t2=&time2);

proc iml worksize=120;
RESET LINESIZE = 80;
%LET faktora = &TIME1;
%LET faktorb = &TIME2;
%LET faktorc = &SUBJECT;

start wald(NN,C,V,p,df,Q,p_val);
Q        = NN * (C * p)` * (ginv(C * V * C`)) * (C * p);
if Q >= 0 then p_val = 1 - probchi(Q,df);
else p_val = .;
finish;

START box(N,p_dach,C,V_d,B1,fg,p);
T        = C` * GINV(C * C`) * C;
tv       = T * V_d;
B1       = (N * p_dach` * T * p_dach) / (trace(tv));
fg       = ((trace(tv))**2)/(trace(tv * tv));
fg2      = 10000;
if B1 >= 0 & fg > 0 then p = 1 - probf(B1,fg,fg2);
else p = .;
FINISH;

START cov (NN,RIJK_,ni,d,V);
/***********************************************************************
    Die Funktion cov berechnet die Kovarianzmatrix des

                          general mixed models

    Die vorliegende Funktion berechnet die Kovarianz des multivariaten
    Modells, Spezialflle wie das CS-Modell mssen extra programmiert
    werden.

    QUELLE: Vorlesungsskript
    'Theory of rank Tests in factorial designs' von Prof. E. Brunner
    15. Oktober 1997.
    Programmiert von:       Andreas Oelerich, AMS Gttingen, Januar 1998

    Die Parameter sind
    NN    : Anzahl der observation
    RIJK_ : die Rnge wobei ber die Messwiederholung gemittelt ist
            (MACRO middle!)
    ni    : Der Vektor der die Anzahl der subjects enthlt
    d     : Anzahl der abhngigen Levels (repeated measures)
    V     : Die Kovarianz die zurckgegeben wird
***********************************************************************/
n = SUM(ni);                        /* Anzahl der subjects            */
r = NROW(ni);                       /* Anzahl der unabhngigen Levels */
pijk_ = 1/NN * (RIJK_ - 1/2);       /* Die im Vektor RIJK_ enthaltenen
                                       Rnge sollen die Mittelrnge
                                       ber die Messwiederholungen der
                                       subjects sein -> MACRO middle!
                                       pijk_ sind die Phi_ijk. des
                                       Papers                         */
IND = LOC(pijk_=.);                 /* wo sind missings?              */
IF NCOL(IND) >0 THEN pijk_[IND] = 0;
pijk_ = SHAPE(pijk_,n,d);           /* ... als Matrix                 */
lijk  = J(n,d,1);                   /* die lambda_ijk des Papers      */
IF NCOL(IND) > 0 THEN lijk[IND] = 0;/* Kodierung !                    */
lp    = lijk # pijk_ ;
DO i=1 TO r;
   IF i=1 THEN DO; r1 = 1;                  r2 = ni[i];        END;
          ELSE DO; r1 = SUM(ni[1:i-1]) + 1; r2 = SUM(ni[1:i]); END;
   lij_   = (lijk[r1:r2,])[+,];     /* ist Vektor!!                   */
   l_1lp  = lp[r1:r2,];
   pij__  = (1 / lij_) # (l_1lp[+,]);
                                    /* Phi_ij..                       */
   vij_d  = J(d,d,0);
   vijj_d = J(d,d,0);
   DO j=1 TO d;
      pij_h = J(ni[i],1,1) @ pij__[j]; /* pij_h ist eine Hilfsmatrix
                                       die die Mittel (pij__)
                                       aufplustert, so ist die
                                       Operation "-" unten zulssig
                                       wegen gleicher Dimension       */
      vij_d[j,j] = SUM(( (lijk[r1:r2,j]) # (pijk_[r1:r2,j] - pij_h))##2) ;
      faktor2    = ni[i] / (lij_[j] * ( lij_[j] - 1 ) );
      vij_d[j,j] = vij_d[j,j] * faktor2;
                                    /* vij_d sind die Schtzer der
                                       Diagonalelemente der Kovarianz-
                                       matrix Vn                      */
      DO j2=1 TO d;
         IF j2 ^= j
            THEN DO;
                   pij_h2 = J(ni[i],1,1) @ pij__[j2];
                   lijj_        = SUM(lijk[r1:r2,j] # lijk[r1:r2,j2]);
                   vijj_d[j,j2] = SUM((lijk[r1:r2,j]) # (lijk[r1:r2,j2])
                    # (pijk_[r1:r2,j] - pij_h) # (pijk_[r1:r2,j2] - pij_h2) );
                   faktor       = ni[i] / ((lij_[j]-1) * (lij_[j2]-1)
                                  + lijj_ - 1);
                   vijj_d[j,j2] = faktor * vijj_d[j,j2];
                 END;               /* of THEN DO                     */
      END;                          /* of DO : j2                     */
   END;                             /* of DO : j                      */
   Vi = n / ni[i] * (vij_d + vijj_d);
   IF i=1 THEN V = Vi;
          ELSE V = BLOCK(Vn,Vi);
END;                                /* of DO : i                      */
FINISH;                             /* of FUNCTION : cov              */

reset nolog;
/********************************************************************/
/*** sortieren des Datensatzes nach den Faktoren A und B          ***/
/*** und lese die Levels der Stufen und die Variable in die       ***/
/*** Vektoren x x_a x_b x_c.                                      ***/
/********************************************************************/


/********************************************************************/
USE _value_;
READ ALL VAR{mijk}     INTO m_ijk;
READ ALL VAR{RmIJK_}   INTO RmIJK;
READ ALL VAR{&subject} INTO pat_;
READ ALL VAR{&time1}   INTO t1_;
READ ALL VAR{&time2}   INTO t2_;
CLOSE _value_;
lev_a = unique(t1_);                  /* Die Stufen des Faktors A ***/
lev_b = unique(t2_);                  /* Die Stufen des Faktors B ***/
lev_c = unique(pat_);                 /* Die Stufen des Faktors B ***/
a     = ncol(lev_a);                  /* Anzahl der Stufen von A  ***/
b     = ncol(lev_b);                  /* Anzahl der Stufen von B  ***/
n     = ncol(lev_c);                  /* Anzahl der subjects      ***/
IF n*a*b ^= nrow(m_ijk)
  THEN DO;
       print 'Error',
             'If you have missing values (empty cell(s))',
             'please fill in one observation with a .';
       END;
m_ijk = SHAPE(m_ijk,n,a*b);
NN    = sum(m_ijk);                   /* Anzahl der Daten           */
/********************************************************************/
/*** i ist der Laufindex des Faktors A  i=1,...,a                 ***/
/*** j ist der Laufindex des Faktors B  j=1,...,b                 ***/
/*** k ist der Laufindex der Subjects   k=1,...,n                 ***/
/*** s ist der Laufindex der Messwiederholungen s=1,...,m_ijk     ***/
/********************************************************************/
/********************************************************************/
/*** Es folgen Definitionen die nicht in jeder Simulations-       ***/
/*** schleife erneut definiert werden muessen.                    ***/
/********************************************************************/
PA    = I(a) - j(a,a,1/a);
PB    = I(b) - j(b,b,1/b);
A1    = j(1,a,1/a);
B1    = j(1,b,1/b);
CA    = PA @ B1;
CB    = A1 @ PB;
CAB   = PA @ PB;
CA_B  = PA   @ I(b);
CB_A  = I(a) @ PB  ;
ranks = RmIJK;
   /*****************************************************************/
   /*** Rijkl    = Vector of rank means for cell (i,j,k)          ***/
   /*** Rijkl_ma = (anxbc)-Matrix der Rangmittelwerte             ***/
   /*** ranks    = R_ijks                                         ***/
   /*** Y_ijks   = H(X_ijks) = (R_ijks-1/2)/NN                    ***/
   /*** p_d      = estimate for relativ treatment effect          ***/
   /*****************************************************************/
   Y_ijkls = (ranks - 1/2) / NN;
   do k=1 to n;
      m_ij = m_ijk[k,];
      mij  = shape(m_ij,a,b);
      if k>1 then do; mij_out = mij_out + mij; end;
             else do; mij_out = mij;           end;
   end;
   Rijk_mat = shape(RmIJK,n,a*b);
   RK_      = Rijk_mat[+,]/n;
   Rij_mat  = shape(RK_,a,b);
   R_J_     = Rij_mat[:,];
   RI__     = Rij_mat[,:];
   R___     = Rij_mat[:,:];
   p_d      = (RK_ - 1/2) / NN ;
   R_vec    = RI__ // (R_J_)` // (RK_)`;
   p        = Y_ijkls;

   r_ijkl = shape(RmIJK,n,a*b);
   /*****************************************************************/
   /*** In der Matrix r_ijk stehen die Rangmittel der Patienten   ***/
   /*****************************************************************/
   r_ijk  = j(n,a*b,0);
   r_ijk  = r_ijkl[:,];
   p_s    = 1/NN *(r_ijk - 1/2);
   p_d    = shape(p_s,a*b,1);

   RUN COV(NN,ranks,n,a*b,V);
   V = V * NN / n;
   /*****************************************************************/
   /*** Die WTS                                                   ***/
   /*** vgl. Agresti & Brunner S. 21                              ***/
   /*****************************************************************/
   Q     = j(5,1,0);
   p_val = j(5,1,0);
   df    = j(5,1,0);
   df[1] =  a-1;
   df[2] =  b-1;
   df[3] = (a-1) * (b-1);
   df[4] = (a-1) *  b   ;
   df[5] =  a    * (b-1);
   /*****************************************************************/
   /*** WALD-TYP-STATISTIKEN                                      ***/
   /*****************************************************************/
   run wald(NN,CA   ,V,p_d,df[1] ,QF,p_v) ; /* Effekt A             */
   Q[1]  = QF; p_val[1]  = p_v;
   run wald(NN,CB   ,V,p_d,df[2] ,QF,p_v) ; /* Effekt B             */
   Q[2]  = QF; p_val[2]  = p_v;
   run wald(NN,CAB  ,V,p_d,df[3] ,QF,p_v) ; /* Effekt AB            */
   Q[3]  = QF; p_val[3]  = p_v;
   run wald(NN,CA_B ,V,p_d,df[4] ,QF,p_v) ; /* Effekt A_B           */
   Q[4]  = QF; p_val[4]  = p_v;
   run wald(NN,CB_A ,V,p_d,df[5] ,QF,p_v) ; /* Effekt B_A           */
   Q[5]  = QF; p_val[5]  = p_v;
   wts = Q || df || p_val ;
/*********************************************************************
   Die BOX-Approximation
*********************************************************************/
   Q_box = j(3,1,0);
   p_box = j(3,1,0);
   dfbox = j(3,1,0);
   RUN box(NN,p_d,CA   ,V,QF_,DF_,p_);          /* Effekt A         */
   Q_box[1]  = QF_; dfbox[1]  = DF_ ; p_box[1]  = p_;
   RUN box(NN,p_d,CB   ,V,QF_,DF_,p_);          /* Effekt B         */
   Q_box[2]  = QF_; dfbox[2]  = DF_ ; p_box[2]  = p_;
   RUN box(NN,p_d,CAB  ,V,QF_,DF_,p_);          /* Effekt AB        */
   Q_box[3]  = QF_; dfbox[3]  = DF_ ; p_box[3]  = p_;
   box = q_box || dfbox || p_box;
   /*****************************************************************/
/********************************************************************/
/***    FUNKTION                                                  ***/
/***       Ausdruck der Class Level Information                   ***/
/********************************************************************/
start O_CLI (a,b,n,m_ijk);
reset center;
class  = {T &faktora, C &faktorb};
levels = a // b ;
print / ;
print 'LD_F2 --- subjects x T x C',
      'T, C: fixed, subjects: random';
print 'SAS-datafile-name:  ' "&data",
      'Response variable:  ' "&var";
print 'Class Level Information' ;
print class levels;
reset noname;
print 'Total number of observations ' (sum(m_ijk)) ' ',
      'Total number of subjects     ' (NROW(m_ijk)) ' ',
      'Number of missing values     ' (NCOL(LOC(m_ijk=0)));
print 'RTE  = Relative Treatment Effects          ',
      'Nobs = Number of Observations (do not count',
      'the repeated measurements within the cells)';
finish ; /* Ende von O_CLI */
/*******************************************************************/
/* nu_char FUNKTION                                                */
/* Diese Funktion schreibt numerische Werte in characters um       */
/* und trimmt diese durch max(...) bzw. trimmt diese, falls        */
/* sie schon characters sind.                                      */
/*******************************************************************/
start nu_char(v,v_neu) ;
if type (v) ='N' then v_neu = char(v,max(int (log10(max(abs(v))))+1));
 else v_neu = trim(v);
finish; /* Ende der Funktion nu_char                                       */
/*******************************************************************/
/*** Ausgabe der rankmeans usw.                                  ***/
/*******************************************************************/
start rmeans(a,b,char1,char2,mij_out,p_d,R_vec);
%let FAK_A = %trim(&FAKTORA) ;          /* Leerzeichen abschneiden */
%let FAK_B = %trim(&FAKTORB) ;          /* Leerzeichen abschneiden */

spalte1 = j(a,1,"&FAK_A") // j(b,1,"&FAK_B")
          // j(a*b,1,"&FAK_A"+'*'+"&FAK_B"+' ') ;
do i=1 to a ;
   do j=1 to b ;
      text = text // (char1[i] + '*' + char2[j] )  ;
   end;
end;
spalte2 = (char1`) // (char2`)  // text ;
source  = concat (spalte1,spalte2);      /* 2 Spalten verschmelzen */
nor     = mij_out[,+] // (mij_out[+,])` // shape(mij_out,a*b,1);
pp      = shape(p_d,a,b);
p       = (pp[,:]) // (pp[:,])` // (p_d) ;
*print /;
print source[c={source}] R_vec[format=6.5 c={"Rank mean"}]
      nor[c={"Nobs"}] p[format=6.5 c={"RTE"}];
finish;
/*******************************************************************/
/*** Ausgabe der Warnungen                                       ***/
/*******************************************************************/
start warning(a,b,n,V);
if a*b > n | det(V) = 0 | min(eigval(V)) < 0 then
  print / '';
reset nocenter;
if a*b > n then
  print 'Warning:',
        'There are less subjects than sub-plot factor level combinations.';
else if det(V)=0 then
  print 'Warning:',
        'Do not use the Wald-type-statistic, because the covariance matrix is singular.';
if min(eigval(V)) < 0 & NCOL(LOC(m_ijk=0)) > 0 then
  print 'Warning:',
        'The estimated covariance matrix is not positive semidefinite due to missing values.';
reset center;
finish;
/*******************************************************************/
/*** Ausgabe der Quadratformen und der p_values                  ***/
/*******************************************************************/
start test_out(wald,box);
source  = {"T" , "C" , "TC" , "T|C" , "C|T"};
source2 =  {"T" , "C" , "TC" };
col1   = {"QF"  "DF"  "p-value"};
print /;
print 'Wald-type-statistic',
      'Approximation for large sample sizes with Chi^2_DF';
wald = wald[1:3,];                          /* ohne simple-effekte */
print wald[r=source2][c=col1][format=6.5];
print 'ANOVA-type-statistic',
      'Box-Approximation for small sample size with Chi^2_df',
      '-estimated degree of freedom';
print box[r=source2][c=col1][format=6.5];
finish;
/*******************************************************************/
dataname = name(&data);
varname  = name(&var);
fa1_name = name(&faktora);
fa2_name = name(&faktorb);
fa3_name = name(&faktorc);

run nu_char(lev_a, leva);
run nu_char(lev_b, levb);
run nu_char(lev_c, levc);
run o_cli(a,b,n,m_ijk);
run rmeans(a,b,leva,levb,mij_out,p_d,R_vec);
run warning(a,b,n,V,m_ijk);
run test_out(wts,box);
quit;
%mend ld_f2;
